# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import logging
import hashlib
import os
import shutil
import struct
import tempfile
from tornado.tcpserver import TCPServer
from .constants import Constants


log = logging.getLogger(__name__)


class UploadError(Exception):
    pass


class FileUpload:
    """
    An ongoing file upload process. After this object is created, it expects the
    following members to be set in the specified order:

    - ``upload.path = 'path/to/file'``
    - ``upload.write(chunk)``
    - ``upload.hash = sha512-hash``
    - ``upload.finish()``

    All errors will be silantly registered, up until the call to
    :meth:`.finish`. That last function call will either return `None` (in case
    of success), or raise an UploadError with a message describing the error
    that occurred in any of the previous steps.

    If any of the file chunks could not be written, for example, the upload will
    continue normally and will raise an :class:`.UploadError` when
    :meth:`.finish` is called with the message "ErrorWritingFile".
    """

    def __init__(self, communication):
        self.communication = communication
        self._path = None
        self.sha = hashlib.sha512()
        self._error = None
        self.committed = False

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        """
        Sets the path, makes sure the designated folder exists and creates a
        lock file to prevent parallell processing of the same file.
        """
        self._path = value
        os.makedirs(os.path.dirname(value), exist_ok=True)
        self.tmp = value + '.tmp'
        try:
            op = next(op
                      for op in self.communication.transaction
                      if isinstance(op, FileUpload) and op.tmp == self.tmp)
            op.abort()
            self.communication.transaction.remove(op)
        except StopIteration:
            pass
        try:
            self.file = open(self.tmp, 'xb')
        except OSError as e:
            log.error(e)
            self.error = 'ErrorOpeningFile'

    def __del__(self):
        """
        Remove temporary file on destruction.
        """
        if self.tmp is None:
            return
        try:
            os.unlink(self.tmp)
        except:
            pass

    def abort(self):
        if self.committed:
            os.unlink(self.path)
            if self.tmp:
                shutil.move(self.tmp, self.path)
        self.error = 'Aborted'

    def prepare(self):
        """
        Prepares the operation to be committed and raises an Exception if a
        commit would fail.
        """
        # test if the target path is writable
        open(self.path, 'ab')

    def commit(self):
        """
        Commits this operation, i.e. moves the temporary file to its rightful
        path.
        """
        if os.path.exists(self.path):
            fd, tmp = tempfile.mkstemp(dir=os.path.dirname(self.path))
            os.close(fd)
            shutil.move(self.path, tmp)
            shutil.move(self.tmp, self.path)
            self.tmp = tmp
        else:
            shutil.move(self.tmp, self.path)
            self.tmp = None
        self.committed = True

    def write(self, chunk):
        """
        Writes a part of the file. See class description for details.
        """
        if self.error:
            return
        try:
            self.file.write(chunk)
            self.sha.update(chunk)
        except OSError:
            self.error = 'ErrorWritingFile'

    def finish(self):
        """
        Wraps up the upload process. See class description for details.
        """
        if self.error is None:
            try:
                self.file.close()
            except OSError:
                self.error = 'ErrorClosingFile'
            else:
                if self.sha.digest() != self.hash:
                    self.error = 'HashMismatch'
        if self.error:
            raise UploadError(self.error)

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, error):
        """
        Removes temporary file on error.
        """
        self.file = None
        self._error = error
        try:
            os.unlink(self.tmp)
        except:
            pass


class Communication:
    """
    The conversation between a client and this server process. See
    :ref:`narrative documentation <netfs_protocol>` for the details of the whole
    communication.
    """

    CHUNK_SIZE = 1024 * 32

    def __init__(self, server, stream):
        self.server = server
        self.stream = stream
        self.transaction = []
        stream.set_close_callback(self.stream_closed)
        self.read_op()

    def stream_closed(self):
        """
        Abort all pending operations if the connection is closed prematurely.
        """
        for op in self.transaction:
            op.abort()

    def read_op(self):
        """
        Read the :ref:`job byte <netfs_protocol>` and call `.handle_op`.
        """
        self.stream.read_bytes(1, self.handle_op)

    def handle_op(self, op_bytes):
        """
        Starts the operation designated by given :ref:`job byte
        <netfs_protocol>`.
        """
        op = struct.unpack('!b', op_bytes)[0]
        if op == Constants.REQ_UPLOAD:
            return self.handle_upload()
        elif op == Constants.REQ_PREPARE:
            return self.handle_prepare()
        elif op == Constants.REQ_COMMIT:
            return self.handle_commit()
        elif op == Constants.REQ_ROLLBACK:
            return self.handle_rollback()
        elif op == Constants.REQ_DOWNLOAD:
            return self.handle_download()
        else:
            log.error('Received bogus request byte %d' % op)
            self.stream.close()

    def handle_prepare(self):
        """
        Handles a ``prepare`` operation. See :ref:`narrative documentation
        <netfs_protocol_prepare>` for details.
        """
        log.debug('prepare')
        try:
            for op in self.transaction:
                op.prepare()
            result = Constants.RESP_OK
        except:
            result = Constants.RESP_ERROR
        result = struct.pack('!b', result)
        self.stream.write(result, self.read_op)

    def handle_rollback(self):
        """
        Handles a ``rollback`` operation. See :ref:`narrative documentation
        <netfs_protocol_rollback>` for details.
        """
        log.debug('rollback')
        for op in self.transaction:
            op.abort()
        self.transaction = []

    def handle_commit(self):
        """
        Handles a ``commit`` operation. See :ref:`narrative documentation
        <netfs_protocol_commit>` for details.
        """
        log.debug('commit')
        for op in self.transaction:
            try:
                op.commit()
            except Exception as e:
                log.debug(e)
                for op in self.transaction:
                    op.abort()
                self.transaction = []
                result = struct.pack('!b', Constants.RESP_ERROR)
                self.stream.write(result, self.read_op)
                return
        self.transaction = []
        result = struct.pack('!b', Constants.RESP_OK)
        self.stream.write(result, self.read_op)

    def handle_upload(self):
        """
        Handles a ``upload`` operation. See :ref:`narrative documentation
        <netfs_protocol_upload>` for details.
        """
        upload = FileUpload(self)
        log.debug('upload')

        def read_name_length(length_bytes):
            length = struct.unpack('!i', length_bytes)[0]
            log.debug('  name length = {}'.format(length))
            self.stream.read_bytes(length, read_name)

        def read_name(name_bytes):
            upload.path = self.get_path(str(name_bytes, 'UTF-8'))
            log.debug('  name = {}'.format(upload.path))
            self.stream.read_bytes(8, read_content_length)

        def read_content_length(length_bytes):
            log.debug('  content length bytes = {}'.format(length_bytes))
            length = struct.unpack('!q', length_bytes)[0]
            log.debug('  content length = {}'.format(length))
            self.stream.read_bytes(length, finished,
                                   streaming_callback=handle_chunk)

        def handle_chunk(chunk):
            log.debug('  ... chunk ...')
            upload.write(chunk)

        def finished(_):
            log.debug('  ... done')
            self.stream.read_bytes(512 // 8, read_hash)

        def read_hash(hash_bytes):
            upload.hash = hash_bytes
            log.debug('  hash = {}'.format(upload.hash))
            try:
                upload.finish()
                log.debug('  all ok!')
                self.transaction.append(upload)
                result = struct.pack('!b', Constants.RESP_OK)
                self.stream.write(result, self.read_op)
            except UploadError as e:
                log.warn(e)
                result = struct.pack('!b', Constants.RESP_ERROR)
                self.stream.write(result, self.read_op)
        self.stream.read_bytes(4, read_name_length)

    def handle_download(self):
        """
        Handles a ``download`` operation. See :ref:`narrative documentation
        <netfs_protocol_upload>` for details.
        """
        log.debug('download')
        path = None
        file = None
        sha = hashlib.sha512()

        def read_name_length(length_bytes):
            length = struct.unpack('!i', length_bytes)[0]
            log.debug('  name length = {}'.format(length))
            self.stream.read_bytes(length, read_name)

        def read_name(name_bytes):
            nonlocal path, file, sha
            name = self.get_path(str(name_bytes, 'UTF-8'))
            log.debug('  name = {}'.format(name))
            path = self.get_path(name)
            if os.path.exists(path + '.tmp'):
                log.debug('  uploading')
                data = struct.pack('!b', Constants.RESP_UPLOADING)
                self.stream.write(data, self.read_op)
                return
            try:
                file = open(path, 'rb')
            except OSError:
                log.debug('  not found')
                data = struct.pack('!b', Constants.RESP_NOTFOUND)
                self.stream.write(data, self.read_op)
                return
            try:
                # seek to end, read file size, go back to beginning of file
                file.seek(0, 2)
                data = struct.pack('!b', Constants.RESP_OK)
                data += struct.pack('!q', file.tell())
                log.debug('  length = {}'.format(file.tell()))
                file.seek(0, 0)
                self.stream.write(data, write_chunk)
            except OSError as e:
                log.debug('  error')
                log.error(e)
                data = struct.pack('!b', Constants.RESP_ERROR)
                self.stream.write(data, self.read_op)

        def write_chunk():
            chunk = file.read(self.CHUNK_SIZE)
            if chunk:
                sha.update(chunk)
                self.stream.write(chunk, write_chunk)
                log.debug('  ... chunk ... ({})'.format(len(chunk)))
                return
            log.debug('  ... done')
            file.close()
            log.debug('  hash = {}'.format(sha.digest()))
            data = sha.digest()
            data += struct.pack('!i', int(os.path.getmtime(path)))
            self.stream.write(data, self.read_op)

        self.stream.read_bytes(4, read_name_length)

    def get_path(self, name):
        """
        Gets the real, absolute path to the file designated by *name*.
        """
        return self.server.get_path(name)


class StorageServer(TCPServer):
    """
    A :class:`tornado.tcpserver.TCPServer` that handles netfs clients.

    The first parameter must be the path to the folder where files should be
    managed. All other arguments are passed to the parent constructor.
    """

    def __init__(self, root, **kwargs):
        self.root = os.path.realpath(root)
        TCPServer.__init__(self, **kwargs)

    def handle_stream(self, stream, address):
        # TODO: the current approach using callbacks to stream functions (i.e.
        # stream.read_bytes(amount, callback)) makes it extremely hard to
        # capture all exceptions raised during the communication.  this function
        # (and the proxy, too) should instead use coroutines:
        # http://www.tornadoweb.org/en/stable/gen.html#tornado.gen.coroutine
        Communication(self, stream)

    def get_path(self, name):
        """
        Gets the real, absolute path to the file designated by *name*.
        """
        path = os.path.join(self.root, name)
        if os.path.commonprefix([path, self.root]) != self.root:
            raise ValueError('Invalid path "%s"' % name)
        return path



# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import logging
import shutil
from ._connection import NetfsConnection
from score.init import (
    init_cache_folder, ConfiguredModule, parse_host_port, parse_bool)
import tempfile


log = logging.getLogger(__name__)
defaults = {
    # note: the 'server' value *must* consist of host and port,
    # otherwise the code in init() might cause an error.
    'server': 'localhost:14000',
    'cachedir': None,
    'deltmpcache': True,
    'ctx.member': 'netfs',
}


def init(confdict, ctx=None):
    """
    Initializes this module acoording to :ref:`our module initialization
    guidelines <module_initialization>` with the following configuration keys:

    :confkey:`server` :faint:`[default=localhost:14000]`
        The server to connect to for all remote operations. Read using the
        generic :func:`score.init.parse_host_port`.

        The special value ``None`` indicates that all remote operations will
        immediately raise an exception. It is still possible to use higher
        level functions like :meth:`put <.ConfiguredNetfsModule.put>` and
        :meth:`get <.ConfiguredNetfsModule.get>` (although :meth:`get
        <.ConfiguredNetfsModule.get>` will raise an exception if the requested
        file is not present in the local folder.)

    :confkey:`cachedir` :faint:`[default=None]`
        A local folder that will hold downloaded files. If this value is
        omitted, the module will create a new temporary folder on demand, that
        will be used as the local folder for this session.

    :confkey:`deltmpcache` :faint:`[default=True]`
        This option is only relevant if the configuration did not contain a
        ``cachedir``. If this value is `True`, any temporary folder that was
        created for this session—as described for the configuration value
        ``cachedir``, above—will be removed when the
        :class:`.ConfiguredNetfsModule` is freed.

        The only use case where you might need this configuration is when you
        want to operate on a temporary folder, but still keep its contents when
        you are done with this module.

    """
    conf = dict(defaults.items())
    conf.update(confdict)
    if conf['server'] in (None, 'None'):
        host, port = None, None
    else:
        host, port = parse_host_port(conf['server'], defaults['server'])
    cachedir = None
    delcache = False
    if conf['cachedir']:
        cachedir = init_cache_folder(conf, 'cachedir')
    else:
        delcache = parse_bool(conf['deltmpcache'])
    c = ConfiguredNetfsModule(host, port, cachedir, delcache)
    c.ctx_conf = ctx
    if ctx and conf['ctx.member'] not in ('None', None):
        ctx.register(conf['ctx.member'], lambda _: c.connect())
    return c


class ConfiguredNetfsModule(ConfiguredModule):
    """
    This module's :class:`configuration class
    <score.init.ConfiguredModule>`.
    """

    def __init__(self, host, port, cachedir, delcache):
        super().__init__(__package__)
        self.host = host
        self.port = port
        self._cachedir = cachedir
        self.delcache = delcache

    def __del__(self):
        if self.delcache and self._cachedir:
            shutil.rmtree(self._cachedir)

    @property
    def cachedir(self):
        if not self._cachedir:
            self._cachedir = tempfile.mkdtemp(prefix='netfs-', suffix='.tmp')
        return self._cachedir

    def connect(self):
        """
        Connects to the configured server and returns a
        :class:`.NetfsConnection`.
        """
        return NetfsConnection(self)



# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import logging
from score.netfs.constants import Constants
import struct
from tornado.tcpserver import TCPServer
from .backend import Backend, NotConnected
from .operation import (DownloadOperation, CommitOperation, PrepareOperation,
                        UploadOperation)


log = logging.getLogger(__name__)


class FrontendCommunication:

    def __init__(self, server, stream):
        self.server = server
        self.stream = stream
        self.stream.set_close_callback(self._stream_closed)
        self.transaction_backends = None
        self.read_op()

    def init_transaction(self, callback):
        if self.transaction_backends is not None:
            callback(self.transaction_backends)
            return
        self.transaction_backends = []

        def check(backend):
            try:
                remaining.remove(backend)
                if not remaining:
                    callback(self.transaction_backends)
            except ValueError:
                pass

        def connected(backend):
            backend.add_close_callback(self._backend_closed)
            self.transaction_backends.append(backend)
            check(backend)

        def failed(backend):
            check(backend)

        remaining = []
        for backend in self.backends:
            remaining.append(backend.transaction(connected, failed))

    def remove_from_transaction(self, backend):
        try:
            backend.send(struct.pack('!b', Constants.REQ_ROLLBACK))
        except NotConnected:
            pass
        try:
            self.transaction_backends.remove(backend)
        except ValueError:
            pass

    def read_op(self):
        if self.stream:
            self.stream.read_bytes(1, self.handle_op)

    def handle_op(self, op_bytes):
        op = struct.unpack('!b', op_bytes)[0]
        if op == Constants.REQ_UPLOAD:
            return UploadOperation(self)
        elif op == Constants.REQ_ROLLBACK:
            return self.handle_rollback_request()
        elif op == Constants.REQ_PREPARE:
            return PrepareOperation(self)
        elif op == Constants.REQ_COMMIT:
            return CommitOperation(self)
        elif op == Constants.REQ_DOWNLOAD:
            return DownloadOperation(self)
        else:
            log.error('Received bogus request byte %d' % op)
            self.terminate()

    def terminate(self):
        if self.stream:
            log.debug('terminate')
            self.stream.close()
            self.stream = None

    def read(self, *args, **kwargs):
        if self.stream:
            self.stream.read_bytes(*args, **kwargs)

    def write(self, *args, **kwargs):
        if self.stream:
            self.stream.write(*args, **kwargs)

    @property
    def backends(self):
        return self.server.backends

    def handle_rollback_request(self):
        log.debug('rollback')
        if self.transaction_backends is None:
            self.read_op()
            return
        data = struct.pack('!b', Constants.REQ_ROLLBACK)
        for backend in self.transaction_backends:
            backend.remove_close_callback(self._backend_closed)
            try:
                backend.send(data)
            except NotConnected:
                pass
        self.transaction_backends = None
        self.read_op()

    def _backend_closed(self, backend):
        if self.transaction_backends is None:
            return
        try:
            self.transaction_backends.remove(backend)
        except ValueError:
            pass

    def _stream_closed(self):
        log.debug('connection closed')
        self.stream = None
        if not self.transaction_backends:
            return
        for backend in self.transaction_backends:
            backend.close()


class ProxyServer(TCPServer):

    def __init__(self, backends, **kwargs):
        self.backends = [Backend(b[0], b[1]) for b in backends]
        TCPServer.__init__(self, **kwargs)

    def handle_stream(self, stream, address):
        FrontendCommunication(self, stream)



# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

from .base import Operation
import struct
from score.netfs.constants import Constants
from score.netfs.proxy.backend import NotConnected


class UploadOperation(Operation):

    def __init__(self, frontend):
        super().__init__(frontend, 'upload')
        self.frontend.init_transaction(self.created_transaction)

    def created_transaction(self, transaction):
        self.transaction = transaction
        for backend in transaction:
            backend.add_close_callback(self._backend_closed)
        self.backends = transaction[:]
        self.distribute(struct.pack('!b', Constants.REQ_UPLOAD))
        self.read(4, self.handle_name_length)

    def distribute(self, data):
        self.log.debug('delegating to {} backends: {}'.
                       format(len(self.backends), data))
        for backend in self.backends:
            try:
                backend.send(data)
            except NotConnected:
                pass

    def handle_name_length(self, length_bytes):
        self.distribute(length_bytes)
        length = struct.unpack('!i', length_bytes)[0]
        self.read(length, self.handle_name)

    def handle_name(self, name_bytes):
        self.distribute(name_bytes)
        self.read(8, self.handle_content_length)

    def handle_content_length(self, length_bytes):
        self.distribute(length_bytes)
        length = struct.unpack('!q', length_bytes)[0]
        self.read(length, self.read_hash, streaming_callback=self.handle_chunk)

    def handle_chunk(self, chunk):
        self.distribute(chunk)

    def read_hash(self, _):
        self.read(512 // 8, self.handle_hash)

    def handle_hash(self, hash_bytes):
        if not self.transaction:
            # not a single backend connection received the upload in full,
            # return error response
            result = struct.pack('!b', Constants.RESP_ERROR)
            self.write(result, self.frontend.read_op)
            return
        self.distribute(hash_bytes)
        for backend in self.transaction:
            backend.read(1, self.create_backend_handler(backend))

    def create_backend_handler(self, backend):
        def handler(b):
            self.handle_backend_response(backend, b)
        return handler

    def handle_backend_response(self, backend, status_bytes):
        backend.remove_close_callback(self._backend_closed)
        self.backends.remove(backend)
        status = struct.unpack('!b', status_bytes)[0]
        self.log.debug('{}: {}'.format(backend, status))
        if status != Constants.RESP_OK:
            self.frontend.remove_from_transaction(backend)
        if self.backends:
            # Not all backends have responded yet
            return
        if not self.transaction:
            result = struct.pack('!b', Constants.RESP_ERROR)
        else:
            result = struct.pack('!b', Constants.RESP_OK)
        self.write(result, self.frontend.read_op)

    def _backend_closed(self, backend):
        self.backends.remove(backend)



import fcntl
import hashlib
import socket
import logging
import os
import shutil
import struct
from ._exceptions import CommitFailed, UploadFailed, DownloadFailed
from .constants import Constants
from transaction.interfaces import IDataManager
from zope.interface import implementer

log = logging.getLogger('score.netfs')


class NetfsConnection:

    CHUNK_SIZE = 1024 * 1024

    def __init__(self, conf):
        self.conf = conf
        if self.conf.host:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.connect((self.conf.host, self.conf.port))
            self.socket.settimeout(None)

    def put(self, path, file, ctx=None, *, move=True):
        """
        Uploads a file with given *path* to the server and moves it into the
        cache folder. The *file* is either a string (denoting a file system path
        to the file), or a :term:`file object`.

        In case the *file* parameter was a string, it is possible to keep that
        original file in place by specifying a falsy value for *move*.
        """
        realpath = os.path.realpath(os.path.join(self.conf.cachedir, path))
        path_prefix = os.path.commonprefix((self.conf.cachedir, realpath))
        if path_prefix != self.conf.cachedir:
            raise ValueError('Invalid path: ' + path)
        dirname = os.path.dirname(realpath)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        if isinstance(file, str):
            if move:
                shutil.move(file, realpath)
            else:
                shutil.copy2(file, realpath)
            file = open(realpath, 'rb')
        else:
            shutil.copyfileobj(file, open(realpath, 'wb'))
        if self.conf.host:
            self.upload(path, file, ctx)

    def get(self, path):
        """
        Returns the local path to a file, downloading it from the server, if it
        does not already exist in the local cache folder.
        """
        realpath = os.path.realpath(os.path.join(self.conf.cachedir, path))
        path_prefix = os.path.commonprefix((self.conf.cachedir, realpath))
        if path_prefix != self.conf.cachedir:
            raise ValueError('Invalid path: ' + path)
        if os.path.exists(realpath):
            return realpath
        tmpfile = realpath + '.tmp'
        dirname = os.path.dirname(realpath)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        file = open(tmpfile, 'wb')
        try:
            fcntl.flock(file, fcntl.LOCK_EX)
            if os.path.exists(realpath):
                # another process downloaded the file
                return realpath
            time = self.download(path, file)
            os.rename(tmpfile, realpath)
            os.utime(realpath, (time, time))
        finally:
            fcntl.flock(file, fcntl.LOCK_UN)
            file.close()
        return realpath

    def upload(self, path, file, ctx=None):
        """
        Puts the contents of given :term:`file object` *file* with given *path*
        onto the server.

        You must call :meth:`.commit` to actually persist the upload. If you are
        using the ctx module, though, you should pass a :term:`context object`
        as *ctx*. This will automatically commit the upload if the transaction
        was successful.
        """
        if self.conf.host is None:
            raise UploadFailed('No server configured')
        if not isinstance(path, bytes):
            path = path.encode('UTF-8')
        data = struct.pack('b', Constants.REQ_UPLOAD)
        data += struct.pack('!i', len(path))
        data += path
        file.seek(0, 2)
        data += struct.pack('!q', file.tell())
        file.seek(0, 0)
        sha = hashlib.sha512()
        self._send(data)
        chunk = file.read(self.CHUNK_SIZE)
        while chunk:
            self._send(chunk)
            sha.update(chunk)
            chunk = file.read(self.CHUNK_SIZE)
        self._send(sha.digest())
        response = struct.unpack('b', self._read(1))[0]
        if response != Constants.RESP_OK:
            raise UploadFailed()
        if ctx:
            _CtxDataManager.join(self, ctx.tx_manager)

    def prepare(self):
        """
        Prepares the current transaction. Raises *CommitFailed* if the server
        responded with an error code.
        """
        if self.conf.host is None:
            return
        self._send(struct.pack('b', Constants.REQ_PREPARE))
        response = struct.unpack('b', self._read(1))[0]
        if response != Constants.RESP_OK:
            raise CommitFailed()

    def commit(self):
        """
        Instructs the server to persist all uploaded files, so that other
        clients can find them.
        """
        if self.conf.host is None:
            return
        self._send(struct.pack('b', Constants.REQ_COMMIT))
        response = struct.unpack('b', self._read(1))[0]
        if response != Constants.RESP_OK:
            raise CommitFailed()

    def rollback(self):
        """
        Sends a rollback command to the server.
        """
        if self.conf.host is None:
            return
        self._send(struct.pack('b', Constants.REQ_ROLLBACK))

    def download(self, path, file, retry=1):
        """
        Downloads the file with given *path* from the server and writes it into
        the :term:`file object` *file*.
        """
        if self.conf.host is None:
            raise DownloadFailed('No server configured')
        if not isinstance(path, bytes):
            path = path.encode('UTF-8')
        data = struct.pack('b', Constants.REQ_DOWNLOAD)
        data += struct.pack('!i', len(path))
        data += path
        self._send(data)
        response = struct.unpack('b', self._read(1))[0]
        if response != Constants.RESP_OK:
            raise DownloadFailed(path)
        length = struct.unpack('!q', self._read(8))[0]
        sha = hashlib.sha512()
        while length:
            chunk_size = min(self.CHUNK_SIZE, length)
            chunk = self._read(chunk_size)
            sha.update(chunk)
            file.write(chunk)
            length -= chunk_size
        hash = self._read(512 // 8)
        if sha.digest() != hash:
            if retry > 0:
                return self.download(path, file, retry - 1)
            raise DownloadFailed(path)
        return struct.unpack('!i', self._read(4))[0]

    def _send(self, data):
        log.debug('sending: {}'.format(data))
        totalsent = 0
        while totalsent < len(data):
            sent = self.socket.send(data[totalsent:])
            if sent == 0:
                raise RuntimeError("socket connection broken")
            totalsent = totalsent + sent

    def _read(self, length):
        chunks = []
        bytes_recd = 0
        while bytes_recd < length:
            chunk = self.socket.recv(min(length - bytes_recd, 2048))
            if chunk == b'':
                raise RuntimeError("socket connection broken")
            chunks.append(chunk)
            bytes_recd = bytes_recd + len(chunk)
        log.debug('receiving {}: {}'.format(length, b''.join(chunks)))
        return b''.join(chunks)


@implementer(IDataManager)
class _CtxDataManager:
    """
    An :interface:`IDataManager <transaction.interfaces.IDataManager>`, which
    will commit all uploaded files at the end of the transaction.
    """

    _instances = []

    @classmethod
    def join(cls, connection, tx_manager):
        tx = tx_manager.get()
        if (connection, tx) not in cls._instances:
            tx.join(cls(connection, tx_manager))

    def __init__(self, connection, tx_manager):
        self.transaction_manager = tx_manager
        self.connection = connection
        self.__class__._instances.append((self.connection, tx_manager.get()))

    def abort(self, transaction):
        self.connection.rollback()
        self.__class__._instances.remove((self.connection, transaction))

    def tpc_begin(self, transaction):
        pass

    def commit(self, transaction):
        pass

    def tpc_vote(self, transaction):
        self.connection.prepare()

    def tpc_finish(self, transaction):
        self.connection.commit()
        self.__class__._instances.remove((self.connection, transaction))

    def tpc_abort(self, transaction):
        self.connection.rollback()
        self.__class__._instances.remove((self.connection, transaction))

    def sortKey(self):
        return 'score.netfs(%d)' % id(self)



# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

from .base import Operation
import random
from score.netfs.proxy.backend import NotConnected
from score.netfs.constants import Constants
import struct


class DownloadOperation(Operation):

    def __init__(self, frontend):
        super().__init__(frontend, 'download')
        self.path = None
        self.backend = None
        self.backends = self.frontend.backends[:]
        self.sent_bytes = 0
        self.read(4, self.read_request_name_length)

    def read_request_name_length(self, length_bytes):
        length = struct.unpack('!i', length_bytes)[0]
        self.read(length, self.read_request_name)

    def read_request_name(self, name_bytes):
        self.path = str(name_bytes, 'UTF-8')
        self.log.debug('path = {}'.format(self.path))
        self.response_attempt()

    def response_attempt(self):
        if not self.backends:
            data = struct.pack('!b', Constants.RESP_ERROR)
            self.write(data, self.frontend.read_op)
            if self.backend:
                self.backend.remove_close_callback(self._backend_closed)
            return
        backend = random.choice(self.backends)
        self.backends.remove(backend)
        try:
            backend.add_close_callback(self._backend_closed)
        except NotConnected:
            return self.response_attempt()
        self.log.debug('chosen: {}'.format(backend))
        self.backend = backend
        data = struct.pack('!b', Constants.REQ_DOWNLOAD)
        data += struct.pack('!i', len(self.path.encode('UTF-8')))
        data += self.path.encode('UTF-8')
        self.skipped_bytes = 0
        self.backend.send(data)
        self.backend.read(1, self.handle_response_status)

    def handle_response_status(self, status_bytes):
        status = struct.unpack('!b', status_bytes)[0]
        if status != Constants.RESP_OK:
            self.response_attempt()
            return
        self.write(status_bytes)
        self.backend.read(8, self.handle_response_size)

    def handle_response_size(self, size_bytes):
        self.write(size_bytes)
        size = struct.unpack('!q', size_bytes)[0]
        self.backend.read(size, self.read_response_hash,
                          streaming_callback=self.handle_response_chunk)

    def handle_response_chunk(self, chunk):
        self.write(chunk)

    def read_response_hash(self, _):
        self.backend.read(512 // 8 + 4, self.handle_response_hash)

    def handle_response_hash(self, hash_bytes):
        self.backend.remove_close_callback(self._backend_closed)
        self.write(hash_bytes)
        self.frontend.read_op()

    def write(self, data, *args, **kwargs):
        diff = self.sent_bytes - self.skipped_bytes
        if diff > 0:
            skip = min(diff, len(data))
            self.skipped_bytes += skip
            data = data[skip:]
        self.skipped_bytes += len(data)
        self.sent_bytes += len(data)
        return super().write(data, *args, **kwargs)

    def _backend_closed(self, backend):
        assert self.backend == backend
        backend.remove_close_callback(self._backend_closed)
        if self.sent_bytes > 0 and not self.backends:
            # Already startend sending data, but can't continue due to lack of
            # working backends. The only remaining option is to terminate the
            # frontend connection.
            self.log.debug('lost last backend connection, terminating')
            return self.frontend.terminate()
        self.log.debug('lost backend connection, retrying')
        self.response_attempt()



# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import click
import logging
import logging.config
import score.netfs as netfs
import score.init
import os.path


log = logging.getLogger(__name__)


def init_logging(logconf):
    if logconf is None:
        return
    conf = score.init.config.parse_config_file(logconf)
    logging.config.fileConfig(conf, disable_existing_loggers=False)


@click.group()
def main():
    """
    Client and server for netfs.
    """


@main.command('serve')
@click.option('-h', '--host', default='0.0.0.0')
@click.option('-p', '--port', default=14000)
@click.option('-l', '--logconf',
              type=click.Path(file_okay=True, dir_okay=False))
@click.argument('folder', type=click.Path(file_okay=False, dir_okay=True))
def serve(folder, host, port, logconf=None):
    init_logging(logconf)
    from tornado.ioloop import IOLoop
    from .server import StorageServer
    try:
        server = StorageServer(folder)
        server.listen(port, address=host)
        IOLoop.instance().start()
        IOLoop.instance().close()
    except Exception as e:
        log.exception(e)
        raise


def read_server_conf(section):
    try:
        folder = section['folder']
    except KeyError:
        raise click.ClickException('No folder configuration found')
    try:
        host = section['host']
    except KeyError:
        host = '0.0.0.0'
    try:
        port = int(section['port'])
    except KeyError:
        port = 14000
    except ValueError:
        raise click.ClickException('Configured port could not be parsed')
    return folder, host, port


@main.command('serve-conf')
@click.argument('conf', type=click.Path(file_okay=True, dir_okay=False))
@click.argument('name', required=False)
def serve_conf(conf, name=None):
    # NOTE: the \b in the following string causes click to print a paragraph
    # verbatim, i.e. without rewrapping its content.
    """
    Start server with config file.

    The configuration file must either contain a section [server] with the
    following keys:

    \b
    - `host` (default: 0.0.0.0)
    - `port` (default: 14000)
    - `folder` (required, no default).

    If an optional NAME is provided, the application will instead look into the
    section [server.NAME], but the expected configuration format does not
    change.
    """
    conf = score.init.parse_config_file(conf)
    if 'loggers' in conf:
        logging.config.fileConfig(conf, disable_existing_loggers=False)
    section_name = 'server'
    if name:
        section_name = 'server-%s' % name
    try:
        section = conf[section_name]
    except KeyError:
        raise click.ClickException('Section "%s" not found in configuration'
                                   % section_name)
    folder, host, port = read_server_conf(section)
    if not os.path.isdir(folder):
        raise click.ClickException('Configured folder (%s) does not exist'
                                   % folder)
    from tornado.ioloop import IOLoop
    from .server import StorageServer
    try:
        server = StorageServer(folder)
        server.listen(port, address=host)
        IOLoop.instance().start()
        IOLoop.instance().close()
    except Exception as e:
        log.exception(e)
        raise


@main.command('proxy')
@click.option('-h', '--host', default='0.0.0.0')
@click.option('-p', '--port', default=14000, type=int)
@click.option('-b', '--backend', multiple=True)
@click.option('-l', '--logconf',
              type=click.Path(file_okay=True, dir_okay=False))
def proxy(host, port, backend, logconf=None):
    init_logging(logconf)
    from tornado.ioloop import IOLoop
    from .proxy import ProxyServer
    backends = []
    for b in backend:
        b = b.split(':')
        b[1] = int(b[1])
        backends.append(b)
    try:
        server = ProxyServer(backends)
        server.listen(port, address=host)
        IOLoop.instance().start()
        IOLoop.instance().close()
    except Exception as e:
        log.exception(e)
        raise


def read_proxy_conf(section):
    host = '0.0.0.0'
    port = 14000
    backends = None
    try:
        host = section['host']
    except KeyError:
        pass
    try:
        port = int(section['port'])
    except KeyError:
        port = 14000
    except ValueError:
        raise click.ClickException('Configured port (%s) could not be parsed'
                                   % section['port'])
    try:
        backend_list = score.init.parse_list(section['backends'])
    except KeyError:
        pass
    else:
        backends = []
        for b in backend_list:
            b = b.split(':')
            b[1] = int(b[1])
            backends.append(b)
    return host, port, backends


@main.command('proxy-conf')
@click.argument('conf', type=click.Path(file_okay=True, dir_okay=False))
def proxy_conf(conf, name=None):
    """
    Start proxy with config file.

    The configuration file may contain a section [proxy] with the following
    keys:

    \b
    - `host` (default: 0.0.0.0)
    - `port` (default: 14000)
    - `backends` (default: see below).

    If no such section is found or the section contains no backends definition,
    the configuration will be done using all other sections in the file, parsing
    all [server] and [server.NAME] sections as described in the command
    serve-conf.
    """
    conf = score.init.parse_config_file(conf)
    if 'loggers' in conf:
        logging.config.fileConfig(conf, disable_existing_loggers=False)
    if 'proxy' in conf:
        host, port, backends = read_proxy_conf(conf['proxy'])
        if backends is None:
            log.debug('No backends configured, browsing all sections')
    else:
        log.debug('No proxy config section, falling back to defaults')
        host = '0.0.0.0'
        port = 14000
        backends = None
    if backends is None:
        backends = []
        for section in conf:
            if section == 'server' or section.startswith('server-'):
                # calling the variables here h and p so they don't clash with
                # the proxy host/port
                _, h, p = read_server_conf(conf[section])
                if (h, p) in backends:
                    log.debug('Ignoring duplicate backend %s:%d' % (h, p))
                    continue
                log.debug('Adding backend %s:%d' % (h, p))
                backends.append((h, p))
    if not backends:
        raise click.ClickException('No backends configured')
    from tornado.ioloop import IOLoop
    from .proxy import ProxyServer
    try:
        log.debug('Starting proxy at %s:%d' % (host, port))
        server = ProxyServer(backends)
        server.listen(port, address=host)
        IOLoop.instance().start()
        IOLoop.instance().close()
    except Exception as e:
        log.exception(e)
        raise


@main.command('download')
@click.option('-h', '--host', default='127.0.0.1')
@click.option('-p', '--port', default=14000, type=int)
@click.option('-l', '--logconf',
              type=click.Path(file_okay=True, dir_okay=False))
@click.argument('path')
@click.argument('file', type=click.File(mode='wb'))
def download(host, port, path, file, logconf=None):
    init_logging(logconf)
    conf = netfs.init({'server': '{}:{}'.format(host, port), 'cachedir': '.'})
    conf.connect().download(path, file)


@main.command('upload')
@click.option('-h', '--host', default='127.0.0.1')
@click.option('-p', '--port', default=14000, type=int)
@click.option('-l', '--logconf',
              type=click.Path(file_okay=True, dir_okay=False))
@click.argument('path')
@click.argument('file', type=click.Path(file_okay=True, dir_okay=False))
def upload(host, port, path, file, logconf=None):
    init_logging(logconf)
    conf = netfs.init({'server': '{}:{}'.format(host, port), 'cachedir': '.'})
    fp = open(file, 'rb')
    conn = conf.connect()
    conn.upload(path, fp)
    conn.commit()

if __name__ == '__main__':
    main()
